import tensorflow as _tf
import so_ml_tools as _soml


def add_batch_to_tensor(x: _tf.Tensor) -> _tf.Tensor:
    """
    Adds a batch size to the given tensor if x = (224, 224, 3) then the result will be (0, 224, 224, 3)

    Args:
        x: The tensor

    Returns:
        A tensor with a batch size of 0.
    """
    return _tf.expand_dims(x, axis=0)


def probability_to_class(y_prob: _tf.Tensor) -> _tf.Tensor:
    """
    See ml.util.label.probability_to_class
    """
    return _soml.util.label.probability_to_class(y_prob=y_prob)


def probability_to_binary(y_prob: _tf.Tensor) -> _tf.Tensor:
    """
    See ml.util.label.probability_to_binary
    """
    return _soml.util.label.probability_to_binary(y_prob=y_prob)


def to_prediction(y_prob: _tf.Tensor) -> _tf.Tensor:
    """
    See ml.util.label.to_prediction
    """
    return _soml.util.label.to_prediction(y_prob=y_prob)



import matplotlib.pyplot as _plt
import seaborn as _sns
import itertools


def timeseries(x, y, labels: list[str] = None, title: str = None, x_label: str = None, y_label: str = None,
               start: int = None, figsize: tuple = (8, 6), label_color='black'):
    fig, ax = _plt.subplots(figsize=figsize)

    color_palette = itertools.cycle(_sns.color_palette(palette="tab10", n_colors=len(x)))
    for idx, _x in enumerate(x):
        # _sns.scatterplot(
        _sns.lineplot(
            x=_x[start:],
            y=y[idx][start:],
            color=next(color_palette),
            # s=7,
            ax=ax,
            label=None if labels is None else labels[idx])

    if x_label:
        _plt.xlabel(x_label)
    if y_label:
        _plt.ylabel(y_label)
    if title:
        _plt.title(title)

    fig.patch.set_alpha(0.0)  # Transparant background

    ax.xaxis.label.set_color(label_color)  # Set color of x-axis label
    ax.tick_params(axis='x', colors=label_color)  # Set color of x-axis ticks.

    ax.yaxis.label.set_color(label_color)  # Set color of y-axis label
    ax.tick_params(axis='y', colors=label_color)  # Set color of y-axis ticks.
    ax.title.set_color(label_color)  # Set color of title

    if labels:
        _plt.legend()

    _plt.show()


import pandas as _pd
import numpy as _np
import matplotlib.pyplot as _plt

from sklearn.feature_selection import SelectKBest as _SelectKBest
from sklearn.feature_selection import mutual_info_classif as _mutual_info_classif
from sklearn.feature_selection import f_classif as _f_classif

from sklearn.feature_selection import mutual_info_regression as _mutual_info_regression


def mutual_information_regression(X: _pd.DataFrame, k='all', y=_pd.DataFrame, seed=42) -> _pd.DataFrame:
    _np.random.seed(seed)
    selector = _SelectKBest(score_func=_mutual_info_regression, k=k)
    selector.fit(X, y.values.ravel())
    return _pd.DataFrame(selector.scores_, index=X.columns, columns=['Score']).sort_values(by='Score', ascending=False)


def mutual_information_regression_plot(X: _pd.DataFrame, y=_pd.DataFrame, seed=42, label_color='black') -> None:
    scores = mutual_information_regression(X=X, y=y, seed=seed)
    _plot(scores=scores, label_color=label_color)


def mutual_information_classification(X: _pd.DataFrame, k='all', y=_pd.DataFrame, seed=42) -> _pd.DataFrame:
    _np.random.seed(seed)
    selector = _SelectKBest(score_func=_mutual_info_classif, k=k)
    selector.fit(X, y.values.ravel())
    return _pd.DataFrame(selector.scores_, index=X.columns, columns=['Score']).sort_values(by='Score', ascending=False)


def mutual_information_classification_plot(X: _pd.DataFrame, y=_pd.DataFrame, seed=42, label_color='black') -> None:
    scores = mutual_information_classification(X=X, y=y, seed=seed)
    _plot(scores=scores, label_color=label_color)


def anova_f_classification(X: _pd.DataFrame, k='all', y=_pd.DataFrame, seed=42) -> _pd.DataFrame:
    _np.random.seed(seed)
    selector = _SelectKBest(score_func=_f_classif, k=k)
    selector.fit(X, y.values.ravel())
    return _pd.DataFrame(selector.scores_, index=X.columns, columns=['Score']).sort_values(by='Score', ascending=False)


def anova_f_classification_plot(X: _pd.DataFrame, y=_pd.DataFrame, seed=42, label_color='black') -> None:
    scores = anova_f_classification(X=X, y=y, seed=seed)
    _plot(scores=scores, label_color=label_color)


def _plot(scores: _pd.DataFrame, label_color):
    fig, ax = _plt.subplots(figsize=(10, 6))
    fig.patch.set_alpha(0.0)  # Transparant background

    y_pos = range(len(scores))
    ax_scores = ax.barh(y_pos, scores.values.ravel(), align='center')
    ax.set_yticks(y_pos)
    ax.set_yticklabels(scores.index.values)

    ax.invert_yaxis()  # labels read top-to-bottom
    ax.set_xlabel('Importance Score')
    ax.set_ylabel('Feature')
    ax.set_title('Feature Importance Scores')
    ax.bar_label(ax_scores, fmt='%.2f')

    ax.xaxis.label.set_color(label_color)  # Set color of x-axis label
    ax.tick_params(axis='x', colors=label_color)  # Set color of x-axis ticks.

    ax.yaxis.label.set_color(label_color)  # Set color of y-axis label
    ax.tick_params(axis='y', colors=label_color)  # Set color of y-axis ticks.
    ax.title.set_color(label_color)  # Set color of title

    # Remove anoying white space at top / bottom.
    _plt.margins(y=0, x=.05)

    _plt.show()


import subprocess as _subprocess
import os as _os


def start_tensorboard(target_dir: str = './logs') -> None:
    """Starts the tensorboard application.

    Args:
        target_dir: the target directory containing the logs
    """
    _subprocess.Popen(["tensorboard", "--logdir", target_dir])
    print("http://localhost:6006")


def stop_tensorboard() -> None:
    """Stops the tensorboard application"""
    _os.system("kill $(ps -e | grep 'tensorboard' | awk '{print $1}')")



import tensorflow as _tf

import so_ml_tools as _soml


def set_trainable_on_layers(model: _tf.keras.Model, layer_names=None, trainable=True) -> None:
    for layer in _soml.tf.model.inspect.collect_layers(model, recursive=True, include_trainable=True, include_non_trainable=True):
        if layer_names is None or layer.name in layer_names:
            layer.trainable = trainable


def set_trainable_on_first_n_layers(model: _tf.keras.Model, n, trainable=True) -> None:
    layers = _soml.tf.model.inspect.collect_layers(model, recursive=True, include_trainable=True, include_non_trainable=True)
    for layer in layers[:n]:
        layer.trainable = trainable


def set_trainable_on_last_n_layers(model: _tf.keras.Model, n, trainable=True) -> None:
    layers = _soml.tf.model.inspect.collect_layers(model, recursive=True, include_trainable=True, include_non_trainable=True)
    for layer in layers[-n:]:
        layer.trainable = trainable



import tensorflow as _tf
import numpy as _np
import so_ml_tools as _soml


def classification_dataset_without_labels(model: _tf.keras.Model, dataset: _tf.data.Dataset) -> _np.ndarray:
    """
    Performs predictions on the given model for all entries in the dataset, will
    return y_prob. Does not work well when dataset is being shuffled.

    :param model: model to perform the predictions on
    :param dataset: dataset containing the data to use for the predictions
    :return: y_true, y_pred
    """
    y_prob = None
    for x in dataset:
        batch_pred = model.predict(x)
        if y_prob is None:
            y_prob = batch_pred
        else:
            y_prob = _np.concatenate((y_prob, batch_pred), axis=0)

    return y_prob


def classification_dataset_with_labels(model: _tf.keras.Model, dataset: _tf.data.Dataset) -> (_np.ndarray, _np.ndarray):
    """
    Performs predictions on the given model for all entries in the dataset, will
    return y_pred and y_true as a tuple. Useful for when the dataset has shuffling enabled
    :param model: model to perform the predictions on
    :param dataset: dataset containing the data to use for the predictions
    :return: A list of ´tuple's containing y_true, y_prob
    """
    y_prob = None
    y_true = None
    for x, y in dataset:
        batch_prob = model.predict(x)
        if y_prob is None:
            y_prob = batch_prob
        else:
            y_prob = _np.concatenate((y_prob, batch_prob), axis=0)

        if y_true is None:
            y_true = y
        else:
            y_true = _np.concatenate((y_true, y), axis=0)
    return y_true, y_prob


def regression(model: _tf.keras.Model, x) -> _np.ndarray:
    """
    Performs predictions on the given model for all entries in the dataset, will
    return y_pred and y_true as a tuple. Usefull for when the dataset has shuffling enabled
    :param model: model to perform the predictions on
    :param x: dataset containing the data to use for the predictions
    :return: A list of ´tuple's containing y_true, y_prob
    """
    y_prob = None
    for values in x:
        batch_prob = model.predict(x=_soml.tf.tensor.add_batch_to_tensor(values))
        if y_prob is None:
            y_prob = batch_prob
        else:
            y_prob = _np.concatenate((y_prob, batch_prob), axis=0)

    return y_prob
